@@ -31,9 +31,13 @@
 
    ____________________________________________________________________________
 
+   CONSTANTS:
+   o  PARSINGTOOLS    dict
+   o  DEFAULT_OPTIONS dict
 
-   o Note            class
-   o AnnotatedString class
+   CLASSES:
+   o  Note            class
+   o  AnnotatedString class
 """
 import copy
 import re
@@ -73,6 +77,34 @@ PARSINGTOOLS["regexes"] = {
 }
 PARSINGTOOLS["shortrange_markers_stopchars"] = (' ', '.', ',', ';')
 
+# default values for .options:
+#
+# (pimydoc)AnnotatedString.options
+# ⋅A dict (str:bool) modifying the way ._extract_notes_from_markers*() methods
+# ⋅work.
+# ⋅
+# ⋅o "authorized short/range markers":
+# ⋅
+# ⋅    - either None if every marker is allowed
+# ⋅    - either a tuple of strings
+# ⋅
+# ⋅o  "un-authorized longrange marker:event": None,
+# ⋅o  "un-authorized shortrange marker:event": None,
+# ⋅
+# ⋅    - None if no method called if an short/long un-authorized marker is read
+# ⋅    - a method whose signature is
+# ⋅           (regex_name=regex_name,
+# ⋅            marker=marker,
+# ⋅            markerspan_in_srcstring=markerspan_in_srcstring)
+# ⋅
+# ⋅            regex_name: "shortrange_marker"
+# ⋅                        "longrange_marker"
+DEFAULT_OPTIONS = {
+    "authorized short/range markers": None,
+    "un-authorized longrange marker:event": None,
+    "un-authorized shortrange marker:event": None,
+    }
+
 
 @dataclass
 class Note:
@@ -155,7 +187,7 @@ class AnnotatedString(dict):
             o (str)source_string          : source string to be read
 
         METHODS:
-            o __init__(self, source_string: str = None, parsingtools=None)
+            o __init__(self, parsingtools=None)
             o __str__(self)
             o _extract_notes_from_markers(self, regex_name)
             o _extract_notes_from_markers__longr(self, marker, markerspan_in_srcstring)
@@ -164,22 +196,21 @@ class AnnotatedString(dict):
             o _extract_notes_from_markers__shrink(self, span_in_srcstring)
             o _get_new_note_fullid(self, regex_name=None, markername=None, suffix="auto")
             o _get_unique_id(self)
-            o _init_from_source_string(self, source_string: str)
             o add_note(self, span, infos: str, note_id=None)
+            o init_from_str(self, source_string: str, remember_source_string: bool = True)
             o initialize_parsingtools_with_reasonable_values(self)
             o improved_str(self, rightpadding=False)
             o normalize_note_spans(self)
     """
     def __init__(self,
-                 source_string: str = None,
-                 parsingtools=None):
+                 parsingtools=None,
+                 options=None):
         """
             AnnotatedString.__init__()
 
             ___________________________________________________________________
 
             ARGUMENTS:
-                o (str)source_string, the string to be read and analysed
                 o (None/dict)parsingtools, a dict of regexes and list of
                   chars allowing to detect markers in the source string.
 
@@ -191,9 +222,36 @@ class AnnotatedString(dict):
                     - by defaut, reasonable values are loaded.
                     - you may want to set other parsingtools values by
                       setting <parsingtools> to a dict {...}.
+
+                o options
+                    (pimydoc)AnnotatedString.options
+                    ⋅A dict (str:bool) modifying the way ._extract_notes_from_markers*() methods
+                    ⋅work.
+                    ⋅
+                    ⋅o "authorized short/range markers":
+                    ⋅
+                    ⋅    - either None if every marker is allowed
+                    ⋅    - either a tuple of strings
+                    ⋅
+                    ⋅o  "un-authorized longrange marker:event": None,
+                    ⋅o  "un-authorized shortrange marker:event": None,
+                    ⋅
+                    ⋅    - None if no method called if an short/long un-authorized marker is read
+                    ⋅    - a method whose signature is
+                    ⋅           (regex_name=regex_name,
+                    ⋅            marker=marker,
+                    ⋅            markerspan_in_srcstring=markerspan_in_srcstring)
+                    ⋅
+                    ⋅            regex_name: "shortrange_marker"
+                    ⋅                        "longrange_marker"
         """
         dict.__init__(self)
 
+        if options:
+            self.options = options
+        else:
+            self.options = copy.deepcopy(DEFAULT_OPTIONS)
+
         if parsingtools:
             self.parsingtools = parsingtools
         else:
@@ -202,11 +260,8 @@ class AnnotatedString(dict):
 
         self.errors = ListOfErrorMessages()
         self._next_unique_id = 0
-        self.source_string = source_string
-        self._source_string = source_string
-
-        if source_string:
-            self._init_from_source_string(source_string)
+        self.source_string = None
+        self._source_string = None
 
     def __str__(self):
         """
@@ -259,7 +314,7 @@ class AnnotatedString(dict):
 
             # ---- <markerspan_in_srcstring> initialization -------------------
             # ---- <markertextspan_in_srcstring> initialization ---------------
-            # ---- <note_id> init ---------------------------------------
+            # ---- <note_id> init ---------------------------------------------
             # ---- <infos> init -----------------------------------------------
             # We set <markerspan_in_srcstring> and <markertextspan_in_srcstring>.
             # * <markerspan_in_srcstring> is the portion of text in .source_string that
@@ -302,7 +357,7 @@ class AnnotatedString(dict):
                  infos) = self._extract_notes_from_markers__shortr(marker,
                                                                    markerspan_in_srcstring)
 
-            # (pimydoc)error::ANNOTATEDTEXT-ERRID001
+            # (pimydoc)error::ANNOTATEDTEXT-ERRORID001
             # ⋅(AnnotatedString)marker refers to an empty string
             # ⋅
             # ⋅Example: "abcdef[V]" isn't a valid AnnotatedString source string since
@@ -311,9 +366,8 @@ class AnnotatedString(dict):
                not loop_stop and \
                markertextspan_in_srcstring[0] == markertextspan_in_srcstring[1]:
                 error = MusaMusaError()
-                error.msgid = "ANNOTATEDTEXT-ERRID001"
-                error.msg = f"[{error.msgid}]" \
-                    f" An error occured while applying regex '{regex}' " \
+                error.msgid = "ANNOTATEDTEXT-ERRORID001"
+                error.msg = f" An error occured while applying regex '{regex}' " \
                     f"to self.source_string '{self.source_string}'. " \
                     f"There's an empty string at position {markertextspan_in_srcstring}."
                 self.errors.append(error)
@@ -335,7 +389,7 @@ class AnnotatedString(dict):
 
         # --- long range markers check : no pending long range markers ----
         for note_id, note in self.items():
-            # (pimydoc)error::ANNOTATEDTEXT-ERRID008
+            # (pimydoc)error::ANNOTATEDTEXT-ERRORID008
             # ⋅(AnnotatedString/long range marker)Pending long range marker
             # ⋅
             # ⋅Example: "abc{g:123}def" isn't a valid AnnotatedString source string since
@@ -346,9 +400,8 @@ class AnnotatedString(dict):
             # ⋅longrange marker marker {P:id2} has no ending marker.
             if not note.span_ok():
                 error = MusaMusaError()
-                error.msgid = "ANNOTATEDTEXT-ERRID008"
-                error.msg = f"[{error.msgid}] " \
-                    f"An error occured while reading " \
+                error.msgid = "ANNOTATEDTEXT-ERRORID008"
+                error.msg = "An error occured while reading " \
                     f"self.source_string '{self.source_string}': " \
                     f"Pending long range marker in note_id '{note_id}' : " \
                     f"note={note}"
@@ -358,7 +411,7 @@ class AnnotatedString(dict):
                                            marker,
                                            markerspan_in_srcstring):
         """
-            _extract_notes_from_markers__longr()
+            AnnotatedString._extract_notes_from_markers__longr()
 
             Submethod of ._extract_notes_from_markers().
 
@@ -385,6 +438,37 @@ class AnnotatedString(dict):
         """
         regex_name = "longrange_marker"
 
+        # ---- authorized marker name ? ---------------------------------------
+        # (pimydoc)AnnotatedString.options
+        # ⋅A dict (str:bool) modifying the way ._extract_notes_from_markers*() methods
+        # ⋅work.
+        # ⋅
+        # ⋅o "authorized short/range markers":
+        # ⋅
+        # ⋅    - either None if every marker is allowed
+        # ⋅    - either a tuple of strings
+        # ⋅
+        # ⋅o  "un-authorized longrange marker:event": None,
+        # ⋅o  "un-authorized shortrange marker:event": None,
+        # ⋅
+        # ⋅    - None if no method called if an short/long un-authorized marker is read
+        # ⋅    - a method whose signature is
+        # ⋅           (regex_name=regex_name,
+        # ⋅            marker=marker,
+        # ⋅            markerspan_in_srcstring=markerspan_in_srcstring)
+        # ⋅
+        # ⋅            regex_name: "shortrange_marker"
+        # ⋅                        "longrange_marker"
+        if self.options["authorized short/range markers"] is not None and \
+           marker.group("markername") not in self.options["authorized short/range markers"] and \
+           self.options["un-authorized longrange marker:event"]:
+            self.options["un-authorized longrange marker:event"](
+                regex_name=regex_name,
+                marker=marker,
+                markerspan_in_srcstring=markerspan_in_srcstring)
+            # returned value: not an error (=don't stop the loop) but nothing to return:
+            return False, False, None, None, None
+
         # ---- marker without endmarker symbol: ---------------------------
         if not marker.group("endmarker"):
 
@@ -424,7 +508,7 @@ class AnnotatedString(dict):
                     targetmarker_id = note_id
                     break
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID007
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID007
         # ⋅(AnnotatedString/long range marker)no matching marker for a marker having
         # ⋅the end symbol
         # ⋅
@@ -441,9 +525,8 @@ class AnnotatedString(dict):
             regex = self.parsingtools["regexes"][regex_name]
 
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID007"
-            error.msg = f"[{error.msgid}] " \
-                f"An error occured while applying regex '{regex}' " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID007"
+            error.msg = f"An error occured while applying regex '{regex}' " \
                 f"to self.source_string '{self.source_string}'. " \
                 f"Long range marker (with an end marker symbol) '{marker}' " \
                 f"doesn't match any previous long range marker with the same name."
@@ -459,7 +542,7 @@ class AnnotatedString(dict):
                                           marker,
                                           markerspan_in_srcstring):
         """
-            _extract_notes_from_markers__ruby()
+            AnnotatedString._extract_notes_from_markers__ruby()
 
             Submethod of ._extract_notes_from_markers().
 
@@ -492,7 +575,7 @@ class AnnotatedString(dict):
         _target = self._source_string[markerspan_in_srcstring[0] -
                                       len(supposed_target):markerspan_in_srcstring[0]]
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID005
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID005
         # ⋅(AnnotatedString/ruby marker)empty target string
         # ⋅
         # ⋅Example: "abcdef❪→comment❫" isn't a valid AnnotatedString source string since
@@ -501,16 +584,15 @@ class AnnotatedString(dict):
             regex = self.parsingtools["regexes"][regex_name]
 
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID005"
-            error.msg = f"[{error.msgid}] " \
-                f"An error occured while applying regex '{regex}' " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID005"
+            error.msg = f"An error occured while applying regex '{regex}' " \
                 f"to self.source_string '{self.source_string}'. " \
                 "Ruby marker contains empty target text."
             self.errors.append(error)
 
             return False, True, None, None, None
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID000
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID000
         # ⋅(AnnotatedString/ruby marker)no matching text before the ruby marker
         # ⋅
         # ⋅Example: "abcdef❪adef→comment" isn't a valid AnnotatedString source string since
@@ -519,9 +601,8 @@ class AnnotatedString(dict):
             regex = self.parsingtools["regexes"][regex_name]
 
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID000"
-            error.msg = f"[{error.msgid}] " \
-                f" An error occured while applying regex '{regex}' " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID000"
+            error.msg = f" An error occured while applying regex '{regex}' " \
                 f"to self.source_string '{self.source_string}'. " \
                 f"Ruby marker contains target text '{supposed_target}' but " \
                 f"just before marker there is '{_target}'."
@@ -539,7 +620,7 @@ class AnnotatedString(dict):
         # ---- infos ------------------------------------------------------
         infos = marker.group("infos")
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID006
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID006
         # ⋅(AnnotatedString/ruby marker)empty infos string
         # ⋅
         # ⋅Example: "abcdef❪def→❫" isn't a valid AnnotatedString source string since
@@ -548,9 +629,8 @@ class AnnotatedString(dict):
             regex = self.parsingtools["regexes"][regex_name]
 
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID006"
-            error.msg = f"[{error.msgid}] " \
-                f" An error occured while applying regex '{regex}' " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID006"
+            error.msg = f" An error occured while applying regex '{regex}' " \
                 f"to self.source_string '{self.source_string}'. " \
                 "Ruby marker contains empty infos " \
                 f"defined for target text '{_target}'."
@@ -565,7 +645,7 @@ class AnnotatedString(dict):
                                             marker,
                                             markerspan_in_srcstring):
         """
-            _extract_notes_from_markers__shortr()
+            AnnotatedString._extract_notes_from_markers__shortr()
 
             Submethod of ._extract_notes_from_markers().
 
@@ -592,6 +672,37 @@ class AnnotatedString(dict):
         """
         regex_name = "shortrange_marker"
 
+        # ---- authorized marker name ? ---------------------------------------
+        # (pimydoc)AnnotatedString.options
+        # ⋅A dict (str:bool) modifying the way ._extract_notes_from_markers*() methods
+        # ⋅work.
+        # ⋅
+        # ⋅o "authorized short/range markers":
+        # ⋅
+        # ⋅    - either None if every marker is allowed
+        # ⋅    - either a tuple of strings
+        # ⋅
+        # ⋅o  "un-authorized longrange marker:event": None,
+        # ⋅o  "un-authorized shortrange marker:event": None,
+        # ⋅
+        # ⋅    - None if no method called if an short/long un-authorized marker is read
+        # ⋅    - a method whose signature is
+        # ⋅           (regex_name=regex_name,
+        # ⋅            marker=marker,
+        # ⋅            markerspan_in_srcstring=markerspan_in_srcstring)
+        # ⋅
+        # ⋅            regex_name: "shortrange_marker"
+        # ⋅                        "longrange_marker"
+        if self.options["authorized short/range markers"] is not None and \
+           marker.group("markername") not in self.options["authorized short/range markers"] and \
+           self.options["un-authorized shortrange marker:event"]:
+            self.options["un-authorized shortrange marker:event"](
+                regex_name=regex_name,
+                marker=marker,
+                markerspan_in_srcstring=markerspan_in_srcstring)
+            # returned value: not an error (=don't stop the loop) but nothing to return:
+            return False, False, None, None, None
+
         # we know that <span_x0> is <markerspan_in_srcstring[1]> : we search <span_x1>
         # e.g. if the string is "What are [bold]you reading ?", we search
         # the span corresponding to "you ".
@@ -730,41 +841,6 @@ class AnnotatedString(dict):
         self._next_unique_id += 1
         return self._next_unique_id - 1
 
-    def _init_from_source_string(self,
-                                 source_string: str):
-        """
-            AnnotatedString._extract_from_source_string()
-
-            Internal method called by .__init__()
-
-            Fill self with Note objects by analysing <source_string>.
-
-            ___________________________________________________________________
-
-            ARGUMENT:
-                o (str)source_string: the source to be analysed
-
-            no RETURNED VALUE
-        """
-        if not self.parsingtools:
-            return
-
-        self.source_string = source_string
-        self._source_string = source_string
-
-        # ---- short range markers -------------------------------------------
-        self._extract_notes_from_markers(regex_name="longrange_marker")
-        self._extract_notes_from_markers(regex_name="shortrange_marker")
-        # N.B. : `ruby_marker` must be called after every marker since it requires
-        #        that ._source_string has been cleaned from every other marker.
-        #
-        #        To understand why, consider the following source string:
-        #             "Ô rage, [V]ô [V]désespoir !❪ô désespoir !→commentaire❫
-        #        The ruby_marker target string is "ô désespoir" which can be read
-        #        only if ._source_string becomes:
-        #             "Ô rage, ô désespoir !❪ô désespoir !→commentaire❫
-        self._extract_notes_from_markers(regex_name="ruby_marker")
-
     def add_note(self,
                  span,
                  infos: str,
@@ -789,61 +865,57 @@ class AnnotatedString(dict):
             note_id = self._get_new_note_fullid()
         new_note = Note(span=span, infos=infos)
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID003
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID003
         # ⋅(AnnotatedString)the span_x0 of the note is an integer inferior to 0.
         # ⋅
         # ⋅The first item in .span (i.e. x0) can't be inferior to 0.
         if span[0] is not None and span[0] < 0:
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID003"
-            error.msg = f"[{error.msgid}] " \
-                f" An error occured while " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID003"
+            error.msg = " An error occured while " \
                 f"adding a note (span={span}; infos='{infos}') " \
                 f"to source_string '{self.source_string}': " \
                 "span[0] is an integer inferior to 0."
             self.errors.append(error)
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID004
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID004
         # ⋅(AnnotatedString)the span_x1 of the note is an integer superior to
         # ⋅                 len(._source_string).
         # ⋅
         # ⋅The last item in .span (i.e. x1) can't be superior to len(._source_string).
         if span[1] is not None and span[1] > len(self._source_string):
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID004"
-            error.msg = f"[{error.msgid}] " \
-                f" An error occured while " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID004"
+            error.msg = " An error occured while " \
                 f"adding a note (span={span}; infos='{infos}') " \
                 f"to source_string '{self.source_string}': " \
                 "span[1] is an integer greater than len(self.source_string), " \
                 f"namely {len(self.source_string)}."
             self.errors.append(error)
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID009
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID009
         # ⋅(AnnotatedString)note.span[0] >= note.span[1]
         # ⋅
         # ⋅The two items of note.span must be so that note.span[0] < note.span[1].
         if span[0] is not None and span[1] is not None and \
            span[0] >= span[1]:
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID009"
-            _msg = f"[{error.msgid}] " \
-                f" An error occured while " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID009"
+            error.msg = " An error occured while " \
                 f"adding a note (span={span}; infos='{infos}') " \
                 f"to source_string '{self.source_string}': " \
                 f"anomaly: span_x0(={span[0]}) is greater or equal to span_x1(={span[0]})."
             self.errors.append(error)
 
-        # (pimydoc)error::ANNOTATEDTEXT-ERRID002
+        # (pimydoc)error::ANNOTATEDTEXT-ERRORID002
         # ⋅(AnnotatedString)duplicate note id
         # ⋅
         # ⋅Example: "[V:g12]abc [V:g12]def" isn't a valid AnnotatedString source string since
         # ⋅"g12" is the name of two different notes.
         if note_id in self:
             error = MusaMusaError()
-            error.msgid = "ANNOTATEDTEXT-ERRID002"
-            error.msg = f"[{error.msgid}] " \
-                " An error occured while " \
+            error.msgid = "ANNOTATEDTEXT-ERRORID002"
+            error.msg = " An error occured while " \
                 f"adding a note (span={span}; infos='{infos}') " \
                 f"to source_string '{self.source_string}': " \
                 f"Duplicate note id '{note_id}'."
@@ -851,6 +923,48 @@ class AnnotatedString(dict):
 
         self[note_id] = new_note
 
+    def init_from_str(self,
+                      source_string: str,
+                      remember_source_string: bool = True):
+        """
+            AnnotatedString.init_from_str()
+
+            Fill self with Note objects by analysing <source_string>.
+
+            ___________________________________________________________________
+
+            ARGUMENTS:
+            o (str)source_string:           the source to be analysed
+            o (bool)remember_source_string: keep in memory .source_string
+                                            and ._source_string
+
+            RETURNED VALUE: self
+        """
+        if not self.parsingtools:
+            return self
+
+        self.source_string = source_string
+        self._source_string = source_string
+
+        # ---- short range markers -------------------------------------------
+        self._extract_notes_from_markers(regex_name="longrange_marker")
+        self._extract_notes_from_markers(regex_name="shortrange_marker")
+        # N.B. : `ruby_marker` must be called after every marker since it requires
+        #        that ._source_string has been cleaned from every other marker.
+        #
+        #        To understand why, consider the following source string:
+        #             "Ô rage, [V]ô [V]désespoir !❪ô désespoir !→commentaire❫
+        #        The ruby_marker target string is "ô désespoir" which can be read
+        #        only if ._source_string becomes:
+        #             "Ô rage, ô désespoir !❪ô désespoir !→commentaire❫
+        self._extract_notes_from_markers(regex_name="ruby_marker")
+
+        if not remember_source_string:
+            self.source_string = None
+            self._source_string = None
+
+        return self
+
     def initialize_parsingtools_with_reasonable_values(self):
         """
             AnnotatedString.initialize_parsingtools_with_reasonable_values()
@@ -915,7 +1029,7 @@ class AnnotatedString(dict):
         res.append(f'\"{self.source_string}\"')
         res.append(f'\"{self._source_string}\"')
 
-        # (pimydoc)AnnotatedString.improved_repr()::how are the markers sorted ?
+        # (pimydoc)AnnotatedString.improved_str()::how are the markers sorted ?
         # ⋅The output sorts the markers along (1) str(span.x0) and along (2) the note_id string.
         # ⋅
         # ⋅By example "{P:id1}[V]{g}a{/g}❪a→comment❫ [N:id2]bc def❪def→comment❫{/P:id1}
